from . import metrics, models, dataset, trainers, zenodo, plotting
from .surgery import scvi_operate, trvae_operate

__author__ = ', '.join([
    'Marco Wagenstetter',
    'Mohsen Naghipourfar',
    'Mohammad Lotfollahi',
    'Sergei Rybakov'
])

__email__ = ', '.join([
    'marco.wagenstetter@tum.de',
    'naghipourfar@ce.sharif.edu',
    'mohammad.lotfollahi@helmholtz-muenchen.de',
    'sergei.rybakov@helmholtz-muenchen.de'
])
